/*
	Copyright 2002-2005 - Cycling '74
	R. Luke DuBois luke@cycling74.com
 */

// updated for new architecture by rld 2/25/02
// tweaked and updated for 64bit and Max SDK 7.0.x by Pawel Janicki (http://paweljanicki.jp) 19.08.2015

/*
 * jit.wavescope does a cheap approximation of a neon glow effect. As cell values approach
 * a luminance value they fade into a specified color. Above that luminance they fade to
 * black (0). Values outside of a tolerance boundary are passed. jit.wavescope can work
 * in an input matrix in either greyscale (mode 0) or color (mode 1).
 *
 */

#include "jit.common.h"


// Our Jitter object instance data
typedef struct _jit_wavescope {
	t_object	ob;
    long		accum;
    long        orientation;
    long        legacy_mode;
} t_jit_wavescope;


// prototypes
BEGIN_USING_C_LINKAGE
t_jit_err		jit_wavescope_init				(void);
t_jit_wavescope	*jit_wavescope_new				(void);
void			jit_wavescope_free				(t_jit_wavescope *x);
t_jit_err		jit_wavescope_matrix_calc		(t_jit_wavescope *x, void *inputs, void *outputs);
void			jit_wavescope_calculate_ndim	(t_jit_wavescope *x, long dim, long *dimsize, long planecount, t_jit_matrix_info *in_minfo, char *bip, t_jit_matrix_info *out_minfo, char *bop);
END_USING_C_LINKAGE


// globals
static void *s_jit_wavescope_class = NULL;


/************************************************************************************/

t_jit_err jit_wavescope_init(void)
{
	long			attrflags = JIT_ATTR_GET_DEFER_LOW | JIT_ATTR_SET_USURP_LOW;
	t_jit_object	*attr;
	t_jit_object	*mop;

	s_jit_wavescope_class = jit_class_new("jit_wavescope", (method)jit_wavescope_new, (method)jit_wavescope_free, sizeof(t_jit_wavescope), 0);

	// add matrix operator (mop)
	mop = (t_jit_object *)jit_object_new(_jit_sym_jit_mop, 1, 1); // args are  num inputs and num outputs
	jit_class_addadornment(s_jit_wavescope_class, mop);

	// add method(s)
	jit_class_addmethod(s_jit_wavescope_class, (method)jit_wavescope_matrix_calc, "matrix_calc", A_CANT, 0);

	// add attribute(s)
    attr = (t_jit_object *)jit_object_new(_jit_sym_jit_attr_offset, "accum", _jit_sym_long, attrflags,
                          (method)0L, (method)0L, calcoffset(t_jit_wavescope, accum));
    jit_class_addattr(s_jit_wavescope_class, attr);
    //CLASS_ATTR_STYLE_LABEL(_jit_map_class,"clip",0,"onoff","Clip Values");

    attr = (t_jit_object *)jit_object_new(_jit_sym_jit_attr_offset, "orientation", _jit_sym_long, attrflags,
                                          (method)0L, (method)0L, calcoffset(t_jit_wavescope, orientation));
    jit_class_addattr(s_jit_wavescope_class, attr);
    //CLASS_ATTR_STYLE_LABEL(_jit_map_class,"clip",0,"onoff","Clip Values");

    attr = (t_jit_object *)jit_object_new(_jit_sym_jit_attr_offset, "legacy_mode", _jit_sym_long, attrflags,
                                          (method)0L, (method)0L, calcoffset(t_jit_wavescope, legacy_mode));
    jit_class_addattr(s_jit_wavescope_class, attr);
    //CLASS_ATTR_STYLE_LABEL(_jit_map_class,"clip",0,"onoff","Clip Values");
    
	// finalize class
	jit_class_register(s_jit_wavescope_class);
	return JIT_ERR_NONE;
}


/************************************************************************************/
// Object Life Cycle

t_jit_wavescope *jit_wavescope_new(void)
{
	t_jit_wavescope	*x = NULL;

	x = (t_jit_wavescope *)jit_object_alloc(s_jit_wavescope_class);
	if (x) {
        x->accum = 25;
        x->orientation = 0;
        x->legacy_mode = 0;
	}
	return x;
}


void jit_wavescope_free(t_jit_wavescope *x)
{
	;	// nothing to free for our simple object
}


/************************************************************************************/
// Methods bound to input/inlets

t_jit_err jit_wavescope_matrix_calc(t_jit_wavescope *x, void *inputs, void *outputs)
{
    t_jit_err err=JIT_ERR_NONE;
    long in_savelock,out_savelock;
    t_jit_matrix_info in_minfo,out_minfo;
    int i,j,k,width_in,width_out,height_in, height_out,planecount;
    char *in_bp,*out_bp;
    void *in_matrix, *out_matrix;
    
    uchar my_temp_1 = 0; uchar my_temp_2 = 0;
    
    in_matrix = jit_object_method(inputs, _jit_sym_getindex, 0);
    out_matrix = jit_object_method(outputs, _jit_sym_getindex, 0);
    
    if (x && in_matrix && out_matrix) {
        
        in_savelock = (long) jit_object_method(in_matrix,_jit_sym_lock,1);
        out_savelock = (long) jit_object_method(out_matrix,_jit_sym_lock,1);
        
        jit_object_method(in_matrix,_jit_sym_getinfo,&in_minfo);
        jit_object_method(out_matrix, _jit_sym_getinfo, &out_minfo);
        
        //compatible types?
        if (in_minfo.type!=_jit_sym_char) {
            err=JIT_ERR_MISMATCH_TYPE;
            goto out;
        }
        
        if(in_minfo.planecount == 1)
        {
            switch(x->orientation)
            {
                case 1:
                    out_minfo.dim[0] = 256;
                    out_minfo.dim[1] = in_minfo.dim[1];
                    out_minfo.planecount = 1;
                    out_minfo.type = in_minfo.type;
                    
                    jit_object_method(out_matrix, _jit_sym_setinfo, &out_minfo);
                    jit_object_method(out_matrix, _jit_sym_getinfo, &out_minfo);
                    //jit_object_method(out_matrix, _jit_sym_setall, 0);
                    
                    jit_object_method(in_matrix,_jit_sym_getdata,&in_bp);
                    jit_object_method(out_matrix,_jit_sym_getdata,&out_bp);
                    
                    if (!in_bp) { err=JIT_ERR_GENERIC; goto out;}
                    if (!out_bp) { err=JIT_ERR_GENERIC; goto out;}
                    
                    width_in = in_minfo.dim[0];
                    width_out = out_minfo.dim[0];
                    height_in = in_minfo.dim[1];
                    height_out = out_minfo.dim[1];
                    
                    if(x->legacy_mode > 0) {
                        for (i=0;i<width_in;i++)
                            for(j=0;j<height_in;j++)
                                *(out_bp+(uchar)*(in_bp+i+j*width_in)+j*width_out) += x->accum;
                    }
                    else {
                        for (i=0;i<width_in;i++)
                            for(j=0;j<height_in;j++) {
                                my_temp_1 = *(out_bp+(uchar)*(in_bp+i+j*width_in)+j*width_out);
                                my_temp_2 = my_temp_1 + x->accum;
                                if(my_temp_2 < my_temp_1) my_temp_2 = 255;
                                *(out_bp+(uchar)*(in_bp+i+j*width_in)+j*width_out) = my_temp_2;
                            }
                        }
                    break;
                case 0:
                    out_minfo.dim[0] = in_minfo.dim[0];
                    out_minfo.dim[1] = 256;
                    out_minfo.planecount = 1;
                    out_minfo.type = in_minfo.type;
                    
                    jit_object_method(out_matrix, _jit_sym_setinfo, &out_minfo);
                    jit_object_method(out_matrix, _jit_sym_getinfo, &out_minfo);
                    //jit_object_method(out_matrix, _jit_sym_setall, 0);
                    
                    jit_object_method(in_matrix,_jit_sym_getdata,&in_bp);
                    jit_object_method(out_matrix,_jit_sym_getdata,&out_bp);
                    
                    if (!in_bp) { err=JIT_ERR_GENERIC; goto out;}
                    if (!out_bp) { err=JIT_ERR_GENERIC; goto out;}
                    
                    width_in = in_minfo.dim[0];
                    width_out = out_minfo.dim[0];
                    height_in = in_minfo.dim[1];
                    height_out = out_minfo.dim[1];
                    
                    if(x->legacy_mode > 0) {
                        for (i=0;i<width_in;i++)
                            for(j=0;j<height_in;j++)
                                *(out_bp+i+(255 - (uchar)*(in_bp+i+j*width_in))*width_out) += x->accum;
                    }
                    else {
                        for (i=0;i<width_in;i++)
                            for(j=0;j<height_in;j++) {
                                my_temp_1 = *(out_bp+i+(255 - (uchar)*(in_bp+i+j*width_in))*width_out);
                                my_temp_2 = my_temp_1 + x->accum;
                                if(my_temp_2 < my_temp_1) my_temp_2 = 255;
                                *(out_bp+i+(255 - (uchar)*(in_bp+i+j*width_in))*width_out) = my_temp_2;
                            }
                    }
                    break;
            }
        } else {
            switch(x->orientation)
            {
                case 1:
                    out_minfo.dim[0] = 256;
                    out_minfo.dim[1] = in_minfo.dim[1];
                    out_minfo.planecount = in_minfo.planecount;
                    out_minfo.type = in_minfo.type;
                    
                    jit_object_method(out_matrix, _jit_sym_setinfo, &out_minfo);
                    jit_object_method(out_matrix, _jit_sym_getinfo, &out_minfo);
                    //jit_object_method(out_matrix, _jit_sym_setall, 0);
                    
                    jit_object_method(in_matrix,_jit_sym_getdata,&in_bp);
                    jit_object_method(out_matrix,_jit_sym_getdata,&out_bp);
                    
                    if (!in_bp) { err=JIT_ERR_GENERIC; goto out;}
                    if (!out_bp) { err=JIT_ERR_GENERIC; goto out;}
                    
                    width_in = in_minfo.dim[0];
                    width_out = out_minfo.dim[0];
                    height_in = in_minfo.dim[1];
                    height_out = out_minfo.dim[1];
                    planecount = in_minfo.planecount;
                    
                    if(x->legacy_mode > 0) {
                        for (i=0;i<width_in;i++)
                            for(j=0;j<height_in;j++)
                                for(k=0;k<in_minfo.planecount;k++)
                                    *(out_bp+(uchar)*(in_bp+i*planecount+j*width_in*planecount+k)*planecount+j*width_out*planecount+k) += x->accum;
                    }
                    else {
                        for (i=0;i<width_in;i++)
                            for(j=0;j<height_in;j++)
                                for(k=0;k<in_minfo.planecount;k++) {
                                    my_temp_1 = *(out_bp+(uchar)*(in_bp+i*planecount+j*width_in*planecount+k)*planecount+j*width_out*planecount+k);
                                    my_temp_2 = my_temp_1 + x->accum;
                                    if(my_temp_2 < my_temp_1) my_temp_2 = 255;
                                    *(out_bp+(uchar)*(in_bp+i*planecount+j*width_in*planecount+k)*planecount+j*width_out*planecount+k) = my_temp_2;
                                }
                    }
                    break;
                case 0:
                    out_minfo.dim[0] = in_minfo.dim[0];
                    out_minfo.dim[1] = 256;
                    out_minfo.planecount = in_minfo.planecount;
                    out_minfo.type = in_minfo.type;
                    
                    jit_object_method(out_matrix, _jit_sym_setinfo, &out_minfo);		
                    jit_object_method(out_matrix, _jit_sym_getinfo, &out_minfo);		
                    //jit_object_method(out_matrix, _jit_sym_setall, 0);
                    
                    jit_object_method(in_matrix,_jit_sym_getdata,&in_bp);
                    jit_object_method(out_matrix,_jit_sym_getdata,&out_bp);
                    
                    if (!in_bp) { err=JIT_ERR_GENERIC; goto out;}
                    if (!out_bp) { err=JIT_ERR_GENERIC; goto out;}
                    
                    width_in = in_minfo.dim[0];
                    width_out = out_minfo.dim[0];
                    height_in = in_minfo.dim[1];
                    height_out = out_minfo.dim[1];
                    planecount = in_minfo.planecount;
                    
                    if(x->legacy_mode > 0) {
                        for (i=0;i<width_in;i++)
                            for(j=0;j<height_in;j++)
                                for(k=0;k<in_minfo.planecount;k++)
                                    *(out_bp+i*planecount+(255 - (uchar)*(in_bp+i*planecount+j*width_in*planecount+k))*width_out*planecount+k) += x->accum;
                    }
                    else {
                        for (i=0;i<width_in;i++)
                            for(j=0;j<height_in;j++)
                                for(k=0;k<in_minfo.planecount;k++) {
                                    my_temp_1 = *(out_bp+i*planecount+(255 - (uchar)*(in_bp+i*planecount+j*width_in*planecount+k))*width_out*planecount+k);
                                    my_temp_2 = my_temp_1 + x->accum;
                                    if(my_temp_2 < my_temp_1) my_temp_2 = 255;
                                    *(out_bp+i*planecount+(255 - (uchar)*(in_bp+i*planecount+j*width_in*planecount+k))*width_out*planecount+k) = my_temp_2;
                                }
                    }
                    break;
            }
        }
    } else {
        return JIT_ERR_INVALID_PTR;
    }
    
out:
    jit_object_method(out_matrix,_jit_sym_lock,out_savelock);
    jit_object_method(in_matrix,_jit_sym_lock,in_savelock);
    return err;
}